﻿using System;
using System.IO;
using System.Text;

namespace TestLibrary
{
    /// <summary>
    /// RTF 어휘 분석기
    /// </summary>
    public class RTFLex
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 리더
        /// </summary>
        private TextReader reader = null;

        /// <summary>
        /// EOF
        /// </summary>
        private const int EOF = -1 ;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - RTFLex(reader)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="reader">리더</param>
        public RTFLex(TextReader reader)
        {
            this.reader = reader;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 토큰 타입 엿보기 - PeekTokenType()

        /// <summary>
        /// 토큰 타입 엿보기
        /// </summary>
        /// <returns>토큰 타입</returns>
        public RTFTokenType PeekTokenType()
        {
            int character = this.reader.Peek();

            while(character == '\r' || character == '\n' || character == '\t' || character == '\0')
            {
                character = this.reader.Read();
                character = this.reader.Peek();
            }

            if(character == EOF)
            {
                return RTFTokenType.EOF;
            }
            else
            {
                switch(character)
                {
                    case '{'  : return RTFTokenType.GroupStart;
                    case '}'  : return RTFTokenType.GroupEnd;
                    case '\\' : return RTFTokenType.Control;
                    default   : return RTFTokenType.Text;
                }
            }
        }

        #endregion
        #region 다음 토큰 읽기 - ReadNextToken()

        /// <summary>
        /// 다음 토큰 읽기
        /// </summary>
        /// <returns>다음 토큰</returns>
        public RTFToken ReadNextToken()
        {
            RTFToken token = new RTFToken();

            int character = this.reader.Read();

            if(character == '\"')
            {
                StringBuilder stringBuilder = new StringBuilder();

                while(true)
                {
                    character = this.reader.Read();

                    if(character < 0)
                    {
                        break;
                    }

                    if(character == '\"')
                    {
                        break;
                    }
                    else
                    {
                        stringBuilder.Append((char)character);
                    }
                }

                token.TokenType = RTFTokenType.Text;
                token.Keyword       = stringBuilder.ToString();

                return token;
            }

            while(character == '\r' || character == '\n' || character == '\t' || character == '\0' )
            {
                character = this.reader.Read();
            }

            if(character != EOF)
            {
                switch(character)
                {
                    case '{' :

                        token.TokenType = RTFTokenType.GroupStart;

                        break;

                    case '}' :

                        token.TokenType = RTFTokenType.GroupEnd;

                        break;

                    case '\\' :

                        ParseKeyword(token);
                        break;

                    default:

                        token.TokenType = RTFTokenType.Text;

                        ParseText(character, token);

                        break;
                }
            }
            else
            {
                token.TokenType = RTFTokenType.EOF;
            }

            return token ;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 키워드 파싱하기 - ParseKeyword(token)

        /// <summary>
        /// 키워드 파싱하기
        /// </summary>
        /// <param name="token">토큰</param>
        private void ParseKeyword(RTFToken token)
        {
            bool extended = false;

            int character = this.reader.Peek();

            if(char.IsLetter((char)character) == false)
            {
                this.reader.Read();

                if(character == '*')
                {
                    token.TokenType = RTFTokenType.Keyword;

                    this.reader.Read();

                    extended = true;

                    goto READ_KEYWORD;
                }
                else if(character == '\\' || character == '{' || character == '}')
                {
                    token.TokenType = RTFTokenType.Text;
                    token.Keyword       = ((char)character).ToString();
                }
                else
                {
                    token.TokenType = RTFTokenType.Control;
                    token.Keyword       = ((char)character).ToString();

                    if(token.Keyword == "\'")
                    {
                        StringBuilder textStringBuilder = new StringBuilder();

                        textStringBuilder.Append((char)this.reader.Read());
                        textStringBuilder.Append((char)this.reader.Read());

                        token.HasParameter = true;
                        token.Parameter    = Convert.ToInt32(textStringBuilder.ToString().ToLower(), 16);

                        if(token.Parameter == 0)
                        {
                            token.Parameter = 0;
                        }
                    }
                }

                return;
            }

READ_KEYWORD :

            StringBuilder keywordStringBuilder = new StringBuilder();

            character = this.reader.Peek();

            while(char.IsLetter((char)character))
            {
                this.reader.Read();

                keywordStringBuilder.Append((char)character);

                character = this.reader.Peek();
            }

            if(extended)
            {
                token.TokenType = RTFTokenType.ExtendedKeyword;
            }
            else
            {
                token.TokenType = RTFTokenType.Keyword;
            }

            token.Keyword = keywordStringBuilder.ToString();

            if(char.IsDigit((char)character) || character == '-')
            {
                token.HasParameter = true;

                bool negative = false;

                if(character == '-')
                {
                    negative = true;

                    this.reader.Read();
                }

                character = this.reader.Peek();

                StringBuilder textStringBuilder = new StringBuilder();

                while(char.IsDigit((char)character))
                {
                    this.reader.Read();

                    textStringBuilder.Append((char)character);

                    character = this.reader.Peek();
                }

                int parameter = Convert.ToInt32(textStringBuilder.ToString());

                if(negative)
                {
                    parameter = - parameter;
                }

                token.Parameter = parameter;
            }

            if(character == ' ')
            {
                this.reader.Read();
            }
        }

        #endregion
        #region 공백 문자 지우기 - ClearWhitespace()

        /// <summary>
        /// 공백 문자 지우기
        /// </summary>
        /// <returns>문자</returns>
        private int ClearWhitespace()
        {
            int character = this.reader.Peek();

            while(character == '\r' || character == '\n' || character == '\t' || character == '\0')
            {
                this.reader.Read();

                character = this.reader.Peek();
            }

            return character;
        }

        #endregion
        #region 텍스트 파싱하기 - ParseText(character, token)

        /// <summary>
        /// 텍스트 파싱하기
        /// </summary>
        /// <param name="character">문자</param>
        /// <param name="token">토큰</param>
        private void ParseText(int character, RTFToken token)
        {
            StringBuilder stringBuilder = new StringBuilder(((char)character).ToString());

            character = ClearWhitespace();

            while(character != '\\' && character != '}' && character != '{' && character != EOF)
            {
                this.reader.Read();

                stringBuilder.Append((char)character);

                character = ClearWhitespace();
            }

            token.Keyword = stringBuilder.ToString();
        }

        #endregion
    }
}