using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace TestLibrary
{
    /// <summary>
    /// RTF 리더
    /// </summary>
    public class RTFReader : IDisposable
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 어휘 분석기
        /// </summary>
        private RTFLex lex = null;

        /// <summary>
        /// 내부 리더
        /// </summary>
        private TextReader innerReader = null;

        /// <summary>
        /// 베이스 스트림
        /// </summary>
        private Stream baseStream = null;

        /// <summary>
        /// 현재 토큰
        /// </summary>
        private RTFToken currentToken = null;

        /// <summary>
        /// 그룹 내 첫번째 토큰 여부
        /// </summary>
        private bool firstTokenInGroup = false;

        /// <summary>
        /// 마지막 토큰
        /// </summary>
        private RTFToken lastToken = null;

        /// <summary>
        /// 레벨
        /// </summary>
        private int level = 0;

        /// <summary>
        /// 토큰 카운트
        /// </summary>
        private int tokenCount = 0;

        /// <summary>
        /// 디폴트 프로세스 이용 가능 여부
        /// </summary>
        private bool enableDefaultProcess = true;

        /// <summary>
        /// 레이어 정보 스택
        /// </summary>
        private Stack<RTFRawLayerInfo> layerInfoStack = new Stack<RTFRawLayerInfo>();

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 내부 리더 - InnerReader

        /// <summary>
        /// 내부 리더
        /// </summary>
        public TextReader InnerReader
        {
            get
            {
                return this.innerReader;
            }
        }

        #endregion
        #region 현재 토큰 - CurrentToken

        /// <summary>
        /// 현재 토큰
        /// </summary>
        public RTFToken CurrentToken
        {
            get
            {
                return this.currentToken;
            }
        }

        #endregion
        #region 현재 토큰 타입 - CurrentTokenType

        /// <summary>
        /// 현재 토큰 타입
        /// </summary>
        public RTFTokenType CurrentTokenType
        {
            get
            {
                if(this.currentToken == null)
                {
                    return RTFTokenType.None;
                }
                else
                {
                    return this.currentToken.TokenType;
                }
            }
        }

        #endregion
        #region 현재 키워드 - CurrentKeyword

        /// <summary>
        /// 현재 키워드
        /// </summary>
        public string CurrentKeyword
        {
            get
            {
                if(this.currentToken == null)
                {
                    return null;
                }
                else
                {
                    return this.currentToken.Keyword;
                }
            }
        }

        #endregion
        #region 현재 매개 변수 소유 여부 - HasCurrentParameter

        /// <summary>
        /// 현재 매개 변수 소유 여부
        /// </summary>
        public bool HasCurrentParameter
        {
            get
            {
                if(this.currentToken == null)
                {
                    return false;
                }
                else
                {
                    return this.currentToken.HasParameter;
                }
            }
        }

        #endregion
        #region 현재 매개 변수 - CurrentParameter

        /// <summary>
        /// 현재 매개 변수
        /// </summary>
        public int CurrentParameter
        {
            get
            {
                if(this.currentToken == null)
                {
                    return 0;
                }
                else
                {
                    return this.currentToken.Parameter;
                }
            }
        }

        #endregion
        #region 컨텐트 위치 - ContentPosition

        /// <summary>
        /// 컨텐트 위치
        /// </summary>
        public int ContentPosition
        {
            get
            {
                if(this.baseStream == null)
                {
                    return 0;
                }
                else
                {
                    return (int)this.baseStream.Position;
                }
            }
        }

        #endregion
        #region 컨텐트 길이 - ContentLength

        /// <summary>
        /// 컨텐트 길이
        /// </summary>
        public int ContentLength
        {
            get
            {
                if(this.baseStream == null)
                {
                    return 0;
                }
                else
                {
                    return (int)this.baseStream.Length;
                }
            }
        }

        #endregion
        #region 그룹 내 첫번째 토큰 여부 - FirstTokenInGroup

        /// <summary>
        /// 그룹 내 첫번째 토큰 여부
        /// </summary>
        public bool FirstTokenInGroup
        {
            get
            {
                return this.firstTokenInGroup;
            }
        }

        #endregion
        #region 마지막 토큰 - LastToken

        /// <summary>
        /// 마지막 토큰
        /// </summary>
        public RTFToken LastToken
        {
            get
            {
                return this.lastToken;
            }
        }

        #endregion
        #region 레벨 - Level

        /// <summary>
        /// 레벨
        /// </summary>
        public int Level
        {
            get
            {
                return this.level;
            }
        }

        #endregion
        #region 토큰 카운트 - TokenCount

        /// <summary>
        /// 토큰 카운트
        /// </summary>
        public int TokenCount
        {
            get
            {
                return this.tokenCount;
            }
            set
            {
                this.tokenCount = value;
            }
        }

        #endregion
        #region 디폴트 프로세스 이용 가능 여부 - EnableDefaultProcess

        /// <summary>
        /// 디폴트 프로세스 이용 가능 여부
        /// </summary>
        public bool EnableDefaultProcess
        {
            get
            {
                return this.enableDefaultProcess;
            }
            set
            {
                this.enableDefaultProcess = value;
            }
        }

        #endregion
        #region 현재 레이어 정보 - CurrentLayerInfo

        /// <summary>
        /// 현재 레이어 정보
        /// </summary>
        public RTFRawLayerInfo CurrentLayerInfo
        {
            get
            {
                if(this.layerInfoStack.Count == 0)
                {
                    this.layerInfoStack.Push(new RTFRawLayerInfo());
                }

                return this.layerInfoStack.Peek();
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - RTFReader()

        /// <summary>
        /// 생성자
        /// </summary>
        public RTFReader()
        {
        }

        #endregion
        #region 생성자 - RTFReader(filePath)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="filePath">파일 경로</param>
        public RTFReader(string filePath)
        {
            LoadRTFFile(filePath);
        }

        #endregion
        #region 생성자 - RTFReader(stream)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="stream">스트림</param>
        public RTFReader(Stream stream)
        {
            StreamReader reader = new StreamReader(stream, Encoding.ASCII);

            LoadReader(reader);

            this.baseStream = stream;
        }

        #endregion
        #region 생성자 - RTFReader(reader)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="reader">텍스트 리더</param>
        public RTFReader(TextReader reader)
        {
            LoadReader(reader);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region RTF 파일 로드하기 - LoadRTFFile(filePath)

        /// <summary>
        /// RTF 파일 로드하기
        /// </summary>
        /// <param name="filePath">파일 경로</param>
        /// <returns>처리 결과</returns>
        public bool LoadRTFFile(string filePath)
        {
            this.currentToken = null;

            if(File.Exists(filePath))
            {
                FileStream stream = new FileStream(filePath, FileMode.Open, FileAccess.Read);

                this.innerReader = new StreamReader(stream, Encoding.ASCII);

                this.baseStream = stream;

                this.lex = new RTFLex(this.innerReader);

                return true;
            }

            return false;
        }

        #endregion
        #region 리더 로드하기 - LoadReader(reader)

        /// <summary>
        /// 리더 로드하기
        /// </summary>
        /// <param name="reader">리더</param>
        /// <returns>처리 결과</returns>
        public bool LoadReader(TextReader reader)
        {
            this.currentToken = null;

            if(reader != null)
            {
                this.innerReader = reader;

                this.lex = new RTFLex(this.innerReader);

                return true ;
            }

            return false;
        }

        #endregion
        #region RTF 텍스트 로드하기 - LoadRTFText(rtf)

        /// <summary>
        /// RTF 텍스트 로드하기
        /// </summary>
        /// <param name="rtf">RTF</param>
        /// <returns>처리 결과</returns>
        public bool LoadRTFText(string rtf)
        {
            this.currentToken = null;

            if(rtf != null && rtf.Length > 3)
            {
                this.innerReader = new StringReader(rtf);

                this.lex = new RTFLex(this.innerReader);

                return true;
            }

            return false;
        }

        #endregion
        #region 토큰 타입 엿보기 - PeekTokenType()

        /// <summary>
        /// 토큰 타입 엿보기
        /// </summary>
        /// <returns>토큰 타입</returns>
        public RTFTokenType PeekTokenType()
        {
            return this.lex.PeekTokenType();
        }

        #endregion
        #region 디폴트 프로세스 처리하기 - DefaultProcess()

        /// <summary>
        /// 디폴트 프로세스 처리하기
        /// </summary>
        public void DefaultProcess()
        {
            if(CurrentToken != null)
            {
                switch(CurrentToken.Keyword)
                {
                    case "uc" :

                        CurrentLayerInfo.UCValue = CurrentParameter;

                        break;
                }
            }
        }

        #endregion
        #region 토큰 읽기 - ReadToken()

        /// <summary>
        /// 토큰 읽기
        /// </summary>
        /// <returns>토큰</returns>
        public RTFToken ReadToken()
        {
            this.firstTokenInGroup = false;

            this.lastToken = this.currentToken;

            if(this.lastToken != null && this.lastToken.TokenType == RTFTokenType.GroupStart)
            {
                this.firstTokenInGroup = true;
            }

            this.currentToken = this.lex.ReadNextToken();

            if(this.currentToken == null || this.currentToken.TokenType == RTFTokenType.EOF)
            {
                this.currentToken = null;

                return null;
            }

            this.tokenCount++;

            if(this.currentToken.TokenType == RTFTokenType.GroupStart)
            {
                if(this.layerInfoStack.Count == 0)
                {
                    this.layerInfoStack.Push(new RTFRawLayerInfo());
                }
                else
                {
                    RTFRawLayerInfo layerInfo = this.layerInfoStack.Peek();

                    this.layerInfoStack.Push(layerInfo.Clone());
                }

                this.level++;
            }
            else if(this.currentToken.TokenType == RTFTokenType.GroupEnd)
            {
                if(this.layerInfoStack.Count > 0)
                {
                    this.layerInfoStack.Pop();
                }

                this.level--;
            }

            if(this.EnableDefaultProcess)
            {
                this.DefaultProcess();
            }

            return this.currentToken;
        }

        #endregion
        #region 끝까지 읽기 - ReadToEndGround()

        /// <summary>
        /// 끝까지 읽기
        /// </summary>
        /// <remarks>
        /// 현재 그룹이 끝날 때까지 데이터를 읽고 무시하고 끝을 유지한다.
        /// </remarks>
        public void ReadToEndGround()
        {
            int level = 0;

            while(true)
            {
                int character = this.innerReader.Peek();

                if(character == -1)
                {
                    break;
                }
                else if(character == '{')
                {
                    level++;
                }
                else if(character == '}')
                {
                    level--;

                    if(level < 0)
                    {
                        break;
                    }
                }

                this.innerReader.Read();
            }
        }

        #endregion
        #region 닫기 - Close()

        /// <summary>
        /// 닫기
        /// </summary>
        public void Close()
        {
            if(this.innerReader != null)
            {
                this.innerReader.Close();

                this.innerReader = null;
            }
        }

        #endregion
        #region 문자열 구하기 - ToString()

        /// <summary>
        /// 문자열 구하기
        /// </summary>
        /// <returns>문자열</returns>
        public override string ToString()
        {
            return $"RTFReader Level:{this.level} {CurrentKeyword}";
        }

        #endregion
        #region 리소스 해제하기 - Dispose()

        /// <summary>
        /// 리소스 해제하기
        /// </summary>
        public void Dispose()
        {
            Close();
        }

        #endregion
    }
}